/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_ASSIGN_HDR
#define EXPRESSO_ASSIGN_HDR

#include "ix.h"
#include <algorithm>

/// Namespace for fast arrays for numerical computation
namespace esso {

  // Array base template declaration
  template<int N, class Type>
  class Array;

  // Promotion template declarations
  template<class Op, class ST>
  struct UPromote;
  template<class Op, class LT, class RT>
  struct BPromote;
  template<class Op, class T0>
  struct RPromote;

  // Template used for calculation in binary operators
  template<class Op, class TL, class TR>
  struct BCalculate;

  // The general iterator type
  template<class Type>
  struct Iterator;

  // The actual array iterator is an array template specialization (!)
  template<int N, class Type>
  class Array<N,Iterator<Type> >;

  // Template magic needed for recursion in 
  // elementwize assignment operator
  template<int K>
  struct Assign {
    template<int N, class DType, class ST>
    static inline void CopyScalar(Array<N,DType> &dst,
                                const ST src, 
				const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::CopyScalar(dst,src,len);
	dst.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class SType>
    static inline void Copy(Array<N,DType> &dst,
                          Array<N,SType> &src, 
			  const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::Copy(dst,src,len);
	dst.Step(K-1);
	src.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
      src.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class ST>
    static inline void AddScalar(Array<N,DType> &dst,
				 const ST src, 
				 const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::AddScalar(dst,src,len);
	dst.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class SType>
    static inline void Add(Array<N,DType> &dst,
			   Array<N,SType> &src, 
			  const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::Add(dst,src,len);
	dst.Step(K-1);
	src.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
      src.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class ST>
    static inline void SubScalar(Array<N,DType> &dst,
                                const ST src, 
				const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::SubScalar(dst,src,len);
	dst.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class SType>
    static inline void Sub(Array<N,DType> &dst,
                          Array<N,SType> &src, 
			  const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::Sub(dst,src,len);
	dst.Step(K-1);
	src.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
      src.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class ST>
    static inline void MulScalar(Array<N,DType> &dst,
                                const ST src, 
				const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::MulScalar(dst,src,len);
	dst.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class SType>
    static inline void Mul(Array<N,DType> &dst,
                          Array<N,SType> &src, 
			  const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::Mul(dst,src,len);
	dst.Step(K-1);
	src.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
      src.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class ST>
    static inline void DivScalar(Array<N,DType> &dst,
                                const ST src, 
				const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::DivScalar(dst,src,len);
	dst.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
    }
    template<int N, class DType, class SType>
    static inline void Div(Array<N,DType> &dst,
                          Array<N,SType> &src, 
			  const Ix<N> &len) {
      for(IXTYPE i=len[K-1]; i!=0; --i) {
	// Note that 'i' is not an array index
        Assign<K-1>::Div(dst,src,len);
	dst.Step(K-1);
	src.Step(K-1);
      }
      dst.Step(K-1,-len[K-1]);
      src.Step(K-1,-len[K-1]);
    }
  };

  template<>
  struct Assign<1> {
    template<int N, class DType, class ST>
    static inline void CopyScalar(Array<N,DType> &dst, 
                                const ST src, 
				const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)=src;
      }
    }
    template<int N, class DType, class SType>
    static inline void Copy(const Array<N,DType> &dst,
                          const Array<N,SType> &src,
			  const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)=src.CGet(i);
      }
    }
    template<int N, class DType, class ST>
    static inline void AddScalar(Array<N,DType> &dst, 
				 const ST src, 
				 const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)+=src;
      }
    }
    template<int N, class DType, class SType>
    static inline void Add(const Array<N,DType> &dst,
			   const Array<N,SType> &src,
			   const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)+=src.CGet(i);
      }
    }
    template<int N, class DType, class ST>
    static inline void SubScalar(Array<N,DType> &dst, 
				 const ST src, 
				 const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)-=src;
      }
    }
    template<int N, class DType, class SType>
    static inline void Sub(const Array<N,DType> &dst,
			   const Array<N,SType> &src,
			   const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)-=src.CGet(i);
      }
    }
    template<int N, class DType, class ST>
    static inline void MulScalar(Array<N,DType> &dst, 
				 const ST src, 
				 const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)*=src;
      }
    }
    template<int N, class DType, class SType>
    static inline void Mul(const Array<N,DType> &dst,
			   const Array<N,SType> &src,
			   const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)*=src.CGet(i);
      }
    }
    template<int N, class DType, class ST>
    static inline void DivScalar(Array<N,DType> &dst, 
				 const ST src, 
				 const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)/=src;
      }
    }
    template<int N, class DType, class SType>
    static inline void Div(const Array<N,DType> &dst,
			   const Array<N,SType> &src,
			   const Ix<N> &len) {
      for(IXTYPE i=len[0]; --i>=0;) {
        dst.Get(i)/=src.CGet(i);
      }
    }
  };

  template<int N, class DType, class SType>
  static inline void BeginAssign(Array<N,DType> &dst, 
				 Array<N,SType> &src,
				 Ix<N> & len) {
    for(DIMT k=0; k<N; k++) {
      IXTYPE l=std::max(dst.L(k),src.L(k));
      len[k]=std::min(dst.U(k),src.U(k))-l;
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(len[k]>0);
      dst.Step(k,l);
      src.Step(k,l);
    }
  }
  template<int N, class DType>
  static inline void BeginAssign(Array<N,DType> &dst,
				 Ix<N> & len) {
    for(DIMT k=0; k<N; k++) {
      IXTYPE l=dst.L(k);
      len[k]=dst.U(k)-l;
      dst.Step(k,l);
    }
  }

  ///@name Array assignment operations
  //@{

  /** Elementwise copy
   * Use this operator to fill an array with a scalar value
   *@arg \a dst Destination array object
   *@arg \a src Source scalar
   *@return reference to destination array */
  template<int N, class DType, class ST>
  inline const Array<N,DType> &
  operator <<=(const Array<N,DType> &dst, const ST & src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    BeginAssign(dstit,len);
    Assign<N>::CopyScalar(dstit,src,len);
    return dst;
  }
  /** Elementwise copy
   *
   * Use this operator for template expression assignment i.e. you
   * may copy a single array into another as
   *
   *\c array1 \c <<= \c array2;
   *
   * or you may compute expressions containing arrays and/or scalars
   * and put the result into an array as
   *
   *\c array1 \c <<= \c array2+array3*scalar1-array1;
   *
   * The assignment is done over \b the \b intersection of the index
   * bounds of all array objects involved. The index bounds of \a dst
   * are unchanged by this operation.
   *
   *@arg \a dst Destination array object
   *@arg \a src Source array object
   *@return reference to destination array
   */
  template<int N, class DType, class SType>
  inline const Array<N,DType> &
  operator <<=(const Array<N,DType> &dst, const Array<N,SType> &src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    typename Array<N,SType>::IteratorType srcit(src);
    BeginAssign(dstit,srcit,len);
    Assign<N>::Copy(dstit,srcit,len);
    return dst;
  }
  /** Addition assignment.
   * Add elements of \a src to elements of \a dst
   * Operates on the intersection of index range for \a src and \a dst
   */
  template<int N, class DType, class ST>
  inline const Array<N,DType> &
  operator +=(const Array<N,DType> &dst, const ST & src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    BeginAssign(dstit,len);
    Assign<N>::AddScalar(dstit,src,len);
    return dst;
  }
  /** Scalar addition assignment.
   * Add scalar \a src to elements of \a dst
   */
  template<int N, class DType, class SType>
  inline const Array<N,DType> &
  operator +=(const Array<N,DType> &dst, const Array<N,SType> &src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    typename Array<N,SType>::IteratorType srcit(src);
    BeginAssign(dstit,srcit,len);
    Assign<N>::Add(dstit,srcit,len);
    return dst;
  }

  /** Subtraction assignment.
   * Subtract elements of \a src from elements of \a dst
   * Operates on the intersection of index range for \a src and \a dst
   */
  template<int N, class DType, class ST>
  inline const Array<N,DType> &
  operator -=(const Array<N,DType> &dst, const ST & src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    BeginAssign(dstit,len);
    Assign<N>::SubScalar(dstit,src,len);
    return dst;
  }
  /** Scalar subtraction assignment.
   * Subtract scalar \a src from elements of \a dst
   */
  template<int N, class DType, class SType>
  inline const Array<N,DType> &
  operator -=(const Array<N,DType> &dst, const Array<N,SType> &src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    typename Array<N,SType>::IteratorType srcit(src);
    BeginAssign(dstit,srcit,len);
    Assign<N>::Sub(dstit,srcit,len);
    return dst;
  }

  /** Multiplication assignment.
   * Multipllies elements of \a src to elements of \a dst
   * Operates on the intersection of index range for \a src and \a dst
   */
  template<int N, class DType, class ST>
  inline const Array<N,DType> &
  operator *=(const Array<N,DType> &dst, const ST & src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    BeginAssign(dstit,len);
    Assign<N>::MulScalar(dstit,src,len);
    return dst;
  }
  /** Scalar multiplication assignment.
   * Multiply scalar \a src to elements of \a dst
   */
  template<int N, class DType, class SType>
  inline const Array<N,DType> &
  operator *=(const Array<N,DType> &dst, const Array<N,SType> &src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    typename Array<N,SType>::IteratorType srcit(src);
    BeginAssign(dstit,srcit,len);
    Assign<N>::Mul(dstit,srcit,len);
    return dst;
  }

  /** Scalar quotient assignment.
   * Divide elements of \a dst with scalar \a src
   */
  template<int N, class DType, class ST>
  inline const Array<N,DType> &
  operator /=(const Array<N,DType> &dst, const ST & src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    BeginAssign(dstit,len);
    Assign<N>::DivScalar(dstit,src,len);
    return dst;
  }
  /** Quotient assignment.
   * Divides elements of \a dst with elements of \a src
   * Operates on the intersection of index range for \a src and \a dst
   */
  template<int N, class DType, class SType>
  inline const Array<N,DType> &
  operator /=(const Array<N,DType> &dst, const Array<N,SType> &src) {
    Ix<N> len;
    typename Array<N,DType>::IteratorType dstit(dst);
    typename Array<N,SType>::IteratorType srcit(src);
    BeginAssign(dstit,srcit,len);
    Assign<N>::Div(dstit,srcit,len);
    return dst;
  }
  //@}
}

#endif
