/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
// Example of using Array<N,T,Fix>

#include "expresso.h"

using namespace esso;

int main() {
   // An index objext is often used with Arrays. It has class name
   // Ix<N> where N is the rank. The following constructors exist.
   
   Ix<3> i1;      // Creates a rank 3 index with undefined initial values
   Ix<3> i2(1);   // Creates a rank 3 index with all values set to 1
   Ix<3> i3(1,2,3); // Creates a rank 3 index with specified values

   // The elements of an index object are manipulated with operator []
   
   i3[0]=10;
   i3[2]=15;

   // The index may be assigned to
   
   i2=i3;  // Copy elements in i3 to i2
   i2=24;  // Copy 24 to all elements in i2

   // Now for the arrays.
   
   // The array class name is Array<N,T,Fix> where N is rank and T is
   // element type. Fix is a class name that specifies this particular
   // Array implementation.
   
   // The default constructor creates an object which is empty.
   // For example, this declaration creates an empty array of
   // rank 2 with element type int
   
   Array<2,Fix<int> > a1;
   
   // The most useful constructor is probably the one that specifies
   // the dimension. For example, this declaration creates an array
   // of rank 3 with element type float and index range 0:4, 0:49, 0:32 
   
   Array<3,Fix<float> > a2(5,50,33);

   // The copy constructor creates an array wich is identical to
   // the source array and points to the same memory area:
   
   Array<3,Fix<float> > a3(a2);

   // Or equivalently
   
   Array<3,Fix<float> > a4=a2;

   // Index objects can be used to create an Array:
   
   Array<3,Fix<float> > a5(i3);

   // Using index objects one can also specify a different starting
   // index for the array than the default (zero). For example,
   // this declaration creates an array of rank 3 with element
   // type float and index range 1:15,10:19,0:3. Notice that the
   // first index objects sets the range length as before, and the
   // second object is the start index

   Array<3,Fix<float> > a6(Ix<3>(15,9,3),Ix<3>(1,10,0));

   // The following declaration gives a const Array object. This
   // doesn't mean that you can't change the elements, only that
   // you can't change anything in the Array object, more about this later.

   const Array<3,Fix<float> > a7(10,10,10);
   // If an array for constant elements is needed it is declared like
   // this

   // Array<3,const float,Fix> a8; (Sorry, not yet implemented)

   // In most of the previous examples, the index range and start values
   // were given as constants in the declarations, but they could
   // also have been unknown integer variables. Sometimes one wants
   // to specify a fixed length however. This is done as below.
   // Here is a declaration of an array with two fixed ranks
   // and one variable:

   Array<3,Fix<float[10][20]> > a9(30);

   // If a partially fixed array with non-default start index is
   // needed it is declared like this: (notice that the size of
   // the length is the rank minus the fixed ranks, while the
   // size of the start-index object is always the rank.

   Array<3,Fix<float[10][20]> > a10(Ix<1>(30),Ix<3>(1,1,1));

   // Enough about declarations.
   // To access an element in an array you use operator ():

   a10(1,5,11)=2;

   // or operator [] with an index object

   a10[i3]=3.14;

   // To get information about index range, use the L(int) and U(int)
   // member functions:

   for(int k=a10.L(2); k<a10.U(2); k++)					
     for(int j=a10.L(1); j<a10.U(1); j++)					
       for(int i=a10.L(0); i<a10.U(0); i++)
	 a10(i,j,k)=0;

   // The Len(int) member function is shorthand for U(int)-L(int)

   int totsize = a10.Len(0)*a10.Len(1)*a10.Len(2);

   // If you want to know the distance in memory between two elements
   // along a specific rank, use Stride(int)

   int stride = a10.Stride(2);

   // The assignment operator makes an array destroy the memory
   // it currently refers to (if no other array refers to it) and start
   // refering to the the same area as the source array

   a5=a7;

   // Assignments are perfectly safe! No memory leaks can occur, ever!
   // Each memory block keeps track of how many arrays are refering
   // to it, and is destroyed automatically if and only if the last array
   // stops refering to it, or is destroyed.

   // An array may also be reinitialized with a new memory block
   // and new index range, using member function Realloc(). The
   // parameter format is the same as with the constructors.

   a5.Dealloc();                   
   a5.Realloc(Ix<3>(10,15,20));    
   a5.Realloc(Ix<3>(10,15,20), Ix<3>(1,1,1));

   // The index range may be shifted without changing the memory
   // area refered to

   a5.Shift(0,5);           // Shifts rank 0 5 steps
   a5.Shift(Ix<3>(1,1,3));  // Shifts all ranks different steps

   // A const Array can not be assigned to, neither can it be Realloc-ed
   // or Shift-ed
   
   return 0;
}
