/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_FLEX_HDR
#define EXPRESSO_FLEX_HDR

#include "debug.h"
#include "tmpltool.h"
#include "multiptr.h"
#include "assign.h"
#include "permutation.h"

namespace esso {

///////////////////////////////////////////////////////////////////////
// Some template magic for use internally by Array<N,T> methods
  namespace BasicInternal {

    template<bool Method>
    class Assign {
    public:
      template<int N, class T0, class TypeS>
      static inline void Do(Array<N,T0> &dest, const Array<N,TypeS> &src) {
        dest.Refer(src);
      }
    };

    template<>
    class Assign<false> {
    public:
      template<int N, class T0, class TypeS>
      static inline void Do(Array<N,T0> &dest, const Array<N,TypeS> &src) {
        dest.Clone(src);
      }
    };
  }

  /// Multidimensional, efficient array class for large data sets
  /** Use this array for efficient numerical computations
   * or other applications where powerful and efficient arrays are
   * needed for large data sets.
   * 
   * If efficiency is crusial, see also the specialization 
   * Array<N,Fix<T> > . It is faster but less flexible (cannot permute
   * dimensions).
   *
   * \internal Other Array implementations may be defined as specializations of this
   * template. To be able to use template expressions, Array objects should
   * have the following members:
   *\li \c T The element type
   *\li \c IteratorType The Array type used as iterator in assignment
   *\li \c L(DIMT) const
   *\li \c U(DIMT) const
   * Arrays that contain data should also have the following members
   *\li \c Stride(DIMT) const
   *\li Origo() const
   *\li Start() const
   */
  template<int N, class T0>
  class Array {
  public:
    friend class Array<N,typename TmplTools::FlipConst<T0>::T>;
     
/////////////////////////////
// typedefs and enums
    typedef T0 T;
    enum { fixrank=0 };
    enum { varrank=N };
    enum { check=0 };
    enum { hasdata=1 };
    typedef T0 Type;
    typedef Array<N,Iterator<T0> > IteratorType;
    ///@name constructor
    ///@{
    /** Default constructor. Initialize array with no data and index bounds
     * zero */
    inline Array(): data(), origo(0) ,l(0) ,u(0) {}
    /** Constructor with lower index bound set to zero
     *@arg len0 number of elements for each dimension */
    inline Array(const Ix<N> &len0);
    /** 
     *@arg len0 number of elements for each dimension 
     *@arg l0 lower index bound */
    inline Array(const Ix<N> &len0, const Ix<N> &l0);
    /** Copy constructor
	Creates a reference to the same memory area as \a src */
    inline Array(const Array &src):
      data(src.data), origo(src.origo), l(src.l), u(src.u), stride(src.stride)
      {}
    /** Copy constructor from another array type object
     * If possible, create a reference to the same memory area as 
     * \c src. If this is not possible, create a new memory area
     * and copy \c src elements */
    template<class TypeS>
    inline Array(const Array<N,TypeS> &src) {
      BasicInternal::Assign<(
          Array<N,TypeS>::hasdata &&
          (TmplTools::IsSame<T,
	                     typename Array<N,TypeS>::T>::check ||
           TmplTools::IsSame<typename TmplTools::UnConst<T>::T,
	                     typename Array<N,TypeS>::T>::check))
	>::Do(*this,src);
    }      
    /** Constructor that creates reference to a slice along dimension \c n
     * from higher-dimension array */
    inline Array(const Array<N+1,Type> &src, DIMT n, IXTYPE i);
    /** 1-dimensional array of given length */
    inline Array(IXTYPE i0):
      data(i0), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==1)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0);
      u[0]=i0;
      stride[0]=1;
    }
    /** 2-dimensional array of given length */
    inline Array(IXTYPE i0, IXTYPE i1):
      data(i0*i1), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==2)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1);
      u[0]=i0;
      u[1]=i1;
      stride[0]=1;
      stride[1]=stride[0]*i0;
    }
    /** 3-dimensional array of given length */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2):
      data(i0*i1*i2), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==3)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2);
      u[0]=i0;
      u[1]=i1;
      u[2]=i2;
      stride[0]=1;
      stride[1]=stride[0]*i0;
      stride[2]=stride[1]*i1;
    }
    /** 4-dimensional array of given length */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3):
      data(i0*i1*i2*i3), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==4)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3);
      u[0]=i0;
      u[1]=i1;
      u[2]=i2;
      u[3]=i3;
      stride[0]=1;
      stride[1]=stride[0]*i0;
      stride[2]=stride[1]*i1;
      stride[3]=stride[2]*i2;
    }
    /** 5-dimensional array of given length */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4):
      data(i0*i1*i2*i3*i4), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==5)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3 && 0<=i4);
      u[0]=i0;
      u[1]=i1;
      u[2]=i2;
      u[3]=i3;
      u[4]=i4;
      stride[0]=1;
      stride[1]=stride[0]*i0;
      stride[2]=stride[1]*i1;
      stride[3]=stride[2]*i2;
      stride[4]=stride[3]*i3;
    }
    /** 6-dimensional array of given length */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4, IXTYPE i5):
      data(i0*i1*i2*i3*i4*i5), origo(data.Data()),l(0) {
      TmplTools::TAssert<(N==6)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3 && 0<=i4 && 0<=i5);
      u[0]=i0;
      u[1]=i1;
      u[2]=i2;
      u[3]=i3;
      u[4]=i4;
      u[5]=i5;
      stride[0]=1;
      stride[1]=stride[0]*i0;
      stride[2]=stride[1]*i1;
      stride[3]=stride[2]*i2;
      stride[4]=stride[3]*i3;
      stride[5]=stride[4]*i4;
    }
    ///@}
    ///@name const
    ///@{
    /** Lower index bounds
     * L(n) is the lowest index allowed on dimension \c n
     */
    inline IXTYPE L(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); return l[n]; 
    }
    /** Upper index bounds
     * U(n) is one more than the largest index allowed on dimension \c n
     */
    inline IXTYPE U(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); return u[n]; 
    }
    /** Convenience function giving the number of array elements along
     * given dimension.
     * Simply put, Len(n) is U(n) - L(n)
     */
    inline IXTYPE Len(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); return u[n]-l[n]; 
    }
    /**\internal Number of elements between each position along 
     * dimension \c n.
     *
     * Mostly for internal use.
     * Stride(n) may be useful in some rare circumstances where someone
     * wishes to use pointer arithmetics on element data.
     */
    inline IXTYPE Stride(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); return stride[n]; 
    }
    /**\internal Pointer to element at index (0,0,...,0)
     *
     * Mostly for internal use. Note that this pointer does
     * not point to a valid element if (0,0,...,0) is outside valid
     * index range. */
    inline T * Origo() const { return origo; }
    /** Pointer to first valid element i.e. element at lowest index bound 
     *
     * Useful for passing pointer to fortran routine, but note that
     * if the dimensions are permuted or reversed the data organization
     * is not compatible with fortran.
     */
    inline T * Start() const { return & operator [](l); }
    /**\internal Pointer to underlying, reference counting data object. Rarely useful */
    inline const MultiPtr<T> &Data() const { return data; }
    /**@return true if this array has no data allocated */
    inline bool IsNull() const { return data.Data()==0; }
    ///@}
    ///@name nonconst
    ///@{
    /** Release this object from it's element data and set index
     * bounds to zero */
    inline bool Dealloc();
    /** Release this object from it's element data and allocate a
     * new data set with given size (lower index bounds zero)
     *@return always returns true */
    inline bool Realloc(const Ix<N> &len0);
    /** Release this object from it's element data and allocate a
     * new data set with given size and given lower index bounds
     *@return always returns true */
    inline bool Realloc(const Ix<N> &len0, const Ix<N> &l0);
    /**\internal Release this object from it's element data and set it to
     * point to the given data with given size and lower index bounds 
     * (rarely useful)
     *@return always returns true */
    inline bool Realloc(const Ix<N> &len0, const Ix<N> &l0,
		       const MultiPtr<T> &data0, T * origo0);
    /** Release this object from it's element data and allocate a
     * new, uninitialized data set with given size and index bounds 
     * identical with those of \c src.
     *@return true on success or false if operation failed */
    template<class TypeS>
    inline bool Realloc(const Array<N,TypeS> &src);
    /** Reinitialize this object as an identical copy of \c src
     * (refering to the same data area as \c src)
     *@return true on success or false if operation failed */
    template<class TypeS>
    inline bool Refer(const Array<N,TypeS> &src);
    /** Same as Clone(*this).
     *@return true on success or false if operation failed */
    inline bool Clone();
    /** Release this object from it's element data and allocate a
     * new data set with index bounds identical to those of \c src.
     * Finally copy the element data from \c src
     *
     * Note that this is the same as Realloc followed by element-wise
     * copy.
     *@return true on success or false if operation failed */
    template<class TypeS>
    inline bool Clone(const Array<N,TypeS> &src);

    /** Create a reference to \c src by setting the stride for dimension
     * 0 to 0, that is, all elements along dimension 0 refer to the same
     * element position. New index bounds for dimension zero are given
     * as arguments. */
    template<class TypeS>
    inline Array & Extend(const Array<N-1,TypeS> &src, IXTYPE l0, IXTYPE u0);

    /** Create a reference to a slice of \c src by cutting along
     * dimension \c n at index \c i */
    template<class TypeS>
    inline Array & Rmdim(const Array<N+1,TypeS> &src, DIMT n, IXTYPE i);
    /** Create an identical reference to \c src data if possible (same as 
     * Refer(src) ), otherwise create a new data area and copy \c src 
     * elements into it (same as Clone(src)) */
    template<class TypeS>
    inline Array & operator =(const Array<N,TypeS> &src) {
      BasicInternal::Assign<(
          Array<N,TypeS>::hasdata &&
          (TmplTools::IsSame<T,typename Array<N,TypeS>::T>::check ||
           TmplTools::IsSame<typename TmplTools::UnConst<T>::T,
	                     typename Array<N,TypeS>::T>::check))
	>::Do(*this,src);
      return *this;
    }
    /** Create an identical reference to \c src data (same as Refer(src)) */
    inline Array & operator =(const Array &src) {
      data=src.data;
      origo=src.origo;
      l=src.l;
      u=src.u;
      stride=src.stride;
      return *this;
    }
    /** Shift data and valid index range along dimension \c n */
    inline Array & Shift(DIMT n, IXTYPE i);
    /** Shift data and valid index range along all dimensions */
    inline Array & Shift(const Ix<N> &ix);
    /** Permute dimensions by swapping two */
    inline Array & Permute(DIMT n1, DIMT n2);
    /** Permute dimensions */
    inline Array & Permute(const Permutation<N> &nx);
    /** Restrict valid index range along dimension \c n 
     *
     * Resulting index range along \c n is the intersection of
     * the given range \c [l0,u0[ and the previous index range \c [L(n),U(n)[
     */
    inline Array & Restrict(DIMT n, IXTYPE l0, IXTYPE u0);
    /** Restrict valid index range along all dimensions */
    inline Array & Restrict(const Ix<N> &l0, const Ix<N> &u0);
    /** Restrict lower index bound along dimension \c n */
    inline Array & RestrictL(DIMT n, IXTYPE l0);
    /** Restrict lower index bound along all dimensions */
    inline Array & RestrictL(const Ix<N> &l0);
    /** Restrict upper index bound along dimension \c n */
    inline Array & RestrictU(DIMT n, IXTYPE u0);
    /** Restrict upper index bound along all dimensions */
    inline Array & RestrictU(const Ix<N> &u0);
    /** Reverse order of elements along dimension \c n and
     * adjust valid index range so that element at (0,0,...,0)
     * is unchanged */
    inline Array & Flip(DIMT n);
    /** Reverse order of elements along all dimensions and
     * adjust valid index range so that element at (0,0,...,0)
     * is unchanged */
    inline Array & Flip(const Ix<N,bool> &r);
    /** Reverse order of elements along dimension \c n.
     * Valid index range is unchanged */
    inline Array & Reverse(DIMT n);
    /** Reverse order of elements along all dimensions,
     * Valid index range is unchanged */
    inline Array & Reverse(const Ix<N,bool> &r);
    /** Set this object to point to every \c s 'th element
     * along dimension \c n */
    inline Array & Restride(DIMT n, IXTYPE s);
    /** Set this object to point to every \c s[n] 'th element
     * along dimension \c n for all \c n (along all dimensions) */
    inline Array & Restride(const Ix<N> &s);
    ///@}    
    /** Get 1-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0) const {
      TmplTools::TAssert<(N==1)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]);
      return *(origo+stride[0]*i0);
    }
    /** Get 2-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1) const {
      TmplTools::TAssert<(N==2)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]);
      return *(origo+stride[1]*i1+
                     stride[0]*i0);
    }
    /** Get 3-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2) const {
      TmplTools::TAssert<(N==3)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]);
      return *(origo+stride[2]*i2+
	             stride[1]*i1+
                     stride[0]*i0);
    }
    /** Get 4-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3) const {
      TmplTools::TAssert<(N==4)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]);
      return *(origo+stride[3]*i3+
	             stride[2]*i2+
	             stride[1]*i1+
                     stride[0]*i0);
    }
    /** Get 5-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4) const {
      TmplTools::TAssert<(N==5)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]
		   && l[4]<=i4 && i4<u[4]);
      return *(origo+stride[4]*i4+
	             stride[3]*i3+
	             stride[2]*i2+
	             stride[1]*i1+
                     stride[0]*i0);
    }
    /** Get 6-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4, IXTYPE i5) const {
      TmplTools::TAssert<(N==6)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]
		   && l[4]<=i4 && i4<u[4]
		   && l[5]<=i5 && i5<u[5]);
      return *(origo+stride[5]*i5+
	             stride[4]*i4+
	             stride[3]*i3+
	             stride[2]*i2+
	             stride[1]*i1+
                     stride[0]*i0);
    }
    /** Get element at given position */
    inline T & operator [](const Ix<N> &ix) const {
#ifdef EXPRESSO_DEBUG_CHECK_INDEXING_RANGE
      for(int k=0; k<N; k++) 
	EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[k]<=ix[k] && ix[k]<u[k]);
#endif
      return *(origo+IndexInternal::Recurse<0,N>::SumProd(stride,ix));
    }

/////////////////////////////
// Private data
  private:
    /** Reference counting data object */
    MultiPtr<T> data;
    /** Pointer to element at index (0,0,...,0) */
    T *origo;
    /** Lower index bound */
    Ix<N> l;
    /** Upper index bound */
    Ix<N> u;
    /** Stride, i.e. number of elements between each index along each
     * dimension. */
    Ix<N> stride;
  };

///////////////////////////////////////////////////////////////////////
// Basic array methods implementation
  template<int N, class T0>
  inline bool
  Array<N,T0>::Dealloc() {
    data=MultiPtr<T>();
    origo=data.Data();
    l=0;
    u=0;
    stride=0;
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,T0>::Clone() {
    return Clone(*this);
  }

  template<int N, class T0>
  inline bool
  Array<N,T0>::Realloc(const Ix<N> &len0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) 
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    data=MultiPtr<T>(IndexInternal::Recurse<0,N>::Prod(len0));
    origo=data.Data();
    l=0;
    for(int n=0; n<N; n++)
      u[n]=len0[n];
    stride[0]=1;
    for(int n=1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1];
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,T0>::Realloc(const Ix<N> &len0, const Ix<N> &l0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    data=MultiPtr<T>(IndexInternal::Recurse<0,N>::Prod(len0));
    origo=data.Data();
    l=l0;
    for(int n=0; n<N; n++)
      u[n]=len0[n]+l[n];
    stride[0]=1;
    for(int n=1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1];
    origo=origo-IndexInternal::Recurse<0,N>::SumProd(stride,l);
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,T0>::Realloc(const Ix<N> &len0, const Ix<N> &l0,
                            const MultiPtr<T> &data0, T *origo0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    data=data0;
    origo=origo0;
    l=l0;
    for(int n=0; n<N; n++)
      u[n]=len0[n]+l[n];
    stride[0]=1;
    for(int n=1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1];
    return true;
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Shift(DIMT n, IXTYPE i) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]+=i;
    u[n]+=i;
    origo-=i*stride[n];
    return *this; 
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Shift(const Ix<N> &ix) {
    for(int n=0; n<N; n++)
      Shift(n,ix[n]);
    return *this;  
  } 

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Reverse(DIMT n) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    origo+=stride[n]*(u[n]+l[n]-1);
    stride[n]*=-1;
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Reverse(const Ix<N,bool> &r) {
    for(int n=0; n<N; n++) {
      if(r[n]) {
	origo+=stride[n]*(u[n]+l[n]-1);
	stride[n]*=-1;
      }
    }
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Flip(DIMT n) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    stride[n]*=-1;
    int tmp=l[n];
    l[n]=1-u[n];
    u[n]=1-tmp;
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Flip(const Ix<N,bool> &r) {
    for(int n=0; n<N; n++) {
      if(r[n]) {
	stride[n]*=-1;
	int tmp=l[n];
	l[n]=1-u[n];
	u[n]=1-tmp;
      }
    }
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::RestrictL(DIMT n, IXTYPE l0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]=std::max(l[n],l0);
    if(l[n]>=u[n]) {
      l[n]=u[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::RestrictL(const Ix<N> &l0) {
    for(int n=0; n<N; n++) {
      l[n]=std::max(l[n],l0[n]);
      if(l[n]>=u[n]) {
	l[n]=u[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::RestrictU(DIMT n, IXTYPE u0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    u[n]=std::min(u[n],u0);
    if(u[n]<=l[n]) {
      u[n]=l[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::RestrictU(const Ix<N> &u0) {
    for(int n=0; n<N; n++) {
      u[n]=std::min(u[n],u0[n]);
      if(u[n]<=l[n]) {
	u[n]=l[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Restrict(DIMT n, IXTYPE l0, IXTYPE u0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]=std::max(l[n],l0);
    u[n]=std::min(u[n],u0);
    if(l[n]>=u[n]) {
      l[n]=u[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Restrict(const Ix<N> &l0, const Ix<N> &u0) {
    for(int n=0; n<N; n++) {
      l[n]=std::max(l[n],l0[n]);
      u[n]=std::min(u[n],u0[n]);
      if(l[n]>=u[n]) {
	l[n]=u[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Permute(DIMT n1, DIMT n2) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n1 && n1<N && 0<=n2 && n2<N);
    IXTYPE tmp;
    tmp=stride[n1];
    stride[n1]=stride[n2];
    stride[n2]=tmp;
    tmp=l[n1];
    l[n1]=l[n2];
    l[n2]=tmp;
    tmp=u[n1];
    u[n1]=u[n2];
    u[n2]=tmp;
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Permute(const Permutation<N> &nx) {
    Ix<N> tmp;
    stride = nx*stride;
    l = nx*l;
    u = nx*u;
    return *this;  
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Restride(DIMT n, IXTYPE s) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    if(l[n]>0) l[n]=(l[n]-1)/s+1; else l[n]=l[n]/s;
    if(u[n]>0) u[n]=(u[n]-1)/s+1; else u[n]=u[n]/s;
    stride[n]*=s;
    return *this;
  }

  template<int N, class T0>
  inline Array<N,T0> &
  Array<N,T0>::Restride(const Ix<N> &sx) {
    for(int n=0; n<N; n++)
      Restride(n,sx[n]);
    return *this;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,T0>::Refer(const Array<N,TypeS> &src) {
    data=src.Data();
    origo=src.Origo();
    for(int n=0; n<N; n++) {
      l[n]=src.L(n);
      u[n]=src.U(n);
      stride[n]=src.Stride(n);
    }
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,T0>::Realloc(const Array<N,TypeS> &src) {
    int totsize=1;    
    for(int n=0; n<N; n++) {
      stride[n]=totsize;
      l[n]=src.L(n);
      u[n]=src.U(n);
      totsize*=u[n]-l[n];
    }
    data=MultiPtr<T>(totsize);
    origo=data.Data()-IndexInternal::Recurse<0,N>::SumProd(stride,l);
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,T0>::Clone(const Array<N,TypeS> &src) {
    Array<N,T0> foo;
    if(!foo.Realloc(src))
      return false;
    foo<<=src;
    *this=foo;
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline Array<N,T0> & 
  Array<N,T0>::Extend(const Array<N-1,TypeS> &src,
		      IXTYPE l0, IXTYPE u0) {
    data=src.Data();
    origo=src.Origo();
    for(int n=0; n<N-1; n++) {
      l[n+1]=src.L(n);
      u[n+1]=src.U(n);
      stride[n+1]=src.Stride(n);
    }
    l[0]=l0;
    u[0]=u0;
    stride[0]=0;
    return *this;
  }

  template<int N, class T0>
  template<class TypeS>
  inline Array<N,T0> & 
  Array<N,T0>::Rmdim(const Array<N+1,TypeS> &src, DIMT dim, IXTYPE i) {
    data=src.Data();
    origo=src.Origo()+src.Stride(dim)*i;
    for(int n=0; n<dim; n++) {
      l[n]=src.L(n);
      u[n]=src.U(n);
      stride[n]=src.Stride(n);
    }
    for(int n=dim+1; n<N+1; n++) {
      l[n-1]=src.L(n);
      u[n-1]=src.U(n);
      stride[n-1]=src.Stride(n);
    }
    return *this;
  }

  template<int N, class T0>
  inline 
  Array<N,T0>::Array(const Ix<N> &len0):
    data(IndexInternal::Recurse<0,N>::Prod(len0)), 
    origo(data.Data()), l(0),u(len0)
  {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<N; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    stride[0]=1;
    for(int n=1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1];
  }

  template<int N, class T0>
  inline 
  Array<N,T0>::Array(const Ix<N> &len0, const Ix<N> &l0):
    data(IndexInternal::Recurse<0,N>::Prod(len0)), 
    origo(data.Data()), l(l0)
  {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<N; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    for(int n=0; n<N; n++)
      u[n]=len0[n]+l[n];
    stride[0]=1;
    for(int n=1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1];
    origo=origo-IndexInternal::Recurse<0,N>::SumProd(stride,l);
  }

  template<int N, class T0>
  inline 
  Array<N,T0>::Array(const Array<N+1,Type> &src, DIMT n, IXTYPE i):
    data(src.Data()), origo(src.Origo()+i*src.Stride(n))
  {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N+1);
    EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(src.L(n)<=i && i<src.U(n));
     
    for(int k=0; k<n; k++) {
      l[k]=src.L(k);
      u[k]=src.U(k);
      stride[k]=src.Stride(k);
    }
    for(int k=n; k<N; k++) {
      l[k]=src.L(k+1);
      u[k]=src.U(k+1);
      stride[k]=src.Stride(k+1);
    }
  }

///////////////////////////////////////////////////////////////////////
// Functions on Basic arrays:

}

#endif
